import json
from typing import Any, List, Tuple
import concurrent.futures
import requests

from tikos.config import VER, BASE_URL_API
from tikos.tikos import (
    AddExtractionFileStream,
    ProcessExtractFile,
    GetGraphStructurePerDoc,
    GenerateGraphPerDoc,
    GetGraph,
    GetGraphRetrievalWithDS,
    BuildSC,
    GetReasoning,
    UploadModel,
    UploadModelConfig,
    UploadModelCaseData,
    ProcessModel)

class TikosClient(object):

    def __init__(self, url: str = "", requestId: str = "", authToken: str = ""):
        if url == "":
            self.url = BASE_URL_API
        else:
            self.url = url
        self.requestId = requestId
        self.authToken = authToken

    def __AddFileBase(self, fn, fs):
        try:
            rtnValAdd = AddExtractionFileStream(url=self.url, requestId=self.requestId, authToken=self.authToken,
                                                fileObj=(fn, fs))
            return fn, rtnValAdd
        except Exception as e:
            return False, False

    def __ProcessFileBase(self, fn, fs):
        try:
            jq_schema = ".[]"

            rtnValAdd = AddExtractionFileStream(url=self.url, requestId=self.requestId, authToken=self.authToken,
                                                fileObj=(fn, fs))

            rtnvalProcess = ProcessExtractFile(
                url=self.url,
                requestId=self.requestId,
                authToken=self.authToken,
                fileName=fn,
                jq=jq_schema,
            )

            return fn, rtnValAdd, rtnvalProcess
        except Exception as e:
            return False, False, False

    '''
        Multithreading supported file processing function
        Accepts: List of filenames and file paths as a tuple
    '''
    def addProcessFiles(self, fileObjs: [(str, str)] = None):
        return_dict = []
        futures = []

        with concurrent.futures.ThreadPoolExecutor() as executor:
            for fileObj in fileObjs:
                fn, fl = fileObj
                fs = open(fl, "rb")

                futures.append(executor.submit(self.__ProcessFileBase, fn=fn, fs=fs))

            for future in concurrent.futures.as_completed(futures):
                return_dict.append(future)

        return return_dict

    '''
        Multithreading supported file addition function
        Accepts: List of filenames and file stream as a tuple
    '''
    def addFileStreams(self, fileObjs: [(str, Any)] = None):

        return_dict = []
        futures = []

        with concurrent.futures.ThreadPoolExecutor() as executor:
            for fileObj in fileObjs:
                fn, fs = fileObj

                futures.append(executor.submit(self.__AddFileBase, fn=fn, fs=fs))

            for future in concurrent.futures.as_completed(futures):
                return_dict.append(future)

        return return_dict

    '''
        Multithreading supported combined file addition and processing function
        Accepts: List of filenames and file stream as a tuple
    '''
    def addProcessFileStreams(self, fileObjs: [(str, Any)] = None):
        return_dict = []
        futures = []

        with concurrent.futures.ThreadPoolExecutor() as executor:
            for fileObj in fileObjs:
                fn, fs = fileObj

                futures.append(executor.submit(self.__ProcessFileBase, fn=fn, fs=fs))

            for future in concurrent.futures.as_completed(futures):
                return_dict.append(future)

        return return_dict

    def __ProcessGetGraphStructureBase(self, fn):
        try:
            rtnValG = GetGraphStructurePerDoc(
                url=self.url,
                requestId=self.requestId,
                authToken=self.authToken,
                refDoc=fn,
                modelid=6,
            )

            return fn, rtnValG
        except Exception as e:
            return False, False

    '''
        Multithreading supported graph structure generation function
        Accepts: List of filenames as contexes
    '''
    def generateGraphStructures(self, contexts: [str] = None):

        return_dict = []
        futures = []

        with concurrent.futures.ThreadPoolExecutor() as executor:
            for context in contexts:
                futures.append(executor.submit(self.__ProcessGetGraphStructureBase, fn=context))

            for future in concurrent.futures.as_completed(futures):
                return_dict.append(future)

        return return_dict

    def __ProcessGenerateGraphBase(self, fn):
        try:
            rtnValGS = GetGraphStructurePerDoc(
                url=self.url,
                requestId=self.requestId,
                authToken=self.authToken,
                refDoc=fn,
                modelid=6,
            )

            status_code, reason, text = rtnValGS

            rtnValG = None
            if (status_code == 200) or (status_code == 201):
                jsonResponse = json.loads(text)
                jsonLLMResponse = str(jsonResponse.get("llmExtraction"))

                rtnValG = GenerateGraphPerDoc(
                    url=self.url,
                    requestId=self.requestId,
                    authToken=self.authToken,
                    fileObj=jsonLLMResponse,
                    refDoc=fn,
                    modelid=6,
                )

            return fn, rtnValGS, rtnValG
        except Exception as e:
            return False, False, False

    '''
        Multithreading supported graph creation function
        Accepts: List of filenames as contexes
    '''
    def createGraph(self, contexts: [str] = None):
        return_dict = []
        futures = []

        with concurrent.futures.ThreadPoolExecutor() as executor:
            for context in contexts:
                futures.append(executor.submit(self.__ProcessGenerateGraphBase, fn=context))

            for future in concurrent.futures.as_completed(futures):
                return_dict.append(future)

        return return_dict

    '''
        Graph structure extraction function
    '''
    def getGraph(self):
        rtnVal = GetGraph(
            url=self.url,
            requestId=self.requestId,
            authToken=self.authToken
        )

        return rtnVal

    '''
        Graph retrieval function
        Accepts: Filenames as context and query
    '''
    def getGraphRetrieval(self, context: str = "", retrieveQuery: str = ""):
        rtnVal = GetGraphRetrievalWithDS(
            url=self.url,
            requestId=self.requestId,
            authToken=self.authToken,
            retrieveQuery=retrieveQuery,
            refdoc=context
        )
        return rtnVal

    '''
        Sequential Collection creation function
        Accepts: Case-Type, Data File name as context, Weight Type and Sequential Collection Config
    '''
    def createSequentialCollection(self, caseType: str = "", context: str = "", wType: str = "BIN", scConfig: str = ""):
        rtnVal = BuildSC(
            url=self.url,
            requestId=self.requestId,
            authToken=self.authToken,
            source=context,
            refCaseType=caseType,
            WType=wType,
            scConfig = scConfig
        )

        return rtnVal

    '''
        Sequential Collection reasoning function 
        Accepts: Case-Type, Data File name as context, problem space case as 
        a JSON object string, Weight Type and Reasoning Type
    '''
    def generateReasoning(self, caseType: str = "", context: str = "", problemSpaceCase: str = "", wType: str = "BIN",
                          rType: str = "CAUSAL", nType:int = 0):

        rtnVal = GetReasoning(
            url=self.url,
            requestId=self.requestId,
            authToken=self.authToken,
            WType=wType,
            refCaseType=caseType,
            psCase=problemSpaceCase,
            refDoc=context,
            RType=rType,
            nType=nType,
        )

        return rtnVal

    '''
        Embedded model Upload function
        Accepts: filenames and file stream as a tuple on model
    '''
    def uploadEmbeddingModel(self, modelObj: (str, Any) = None, modelType: str = "DNN"):
        rtnValAdd = UploadModel(url=self.url, requestId=self.requestId, authToken=self.authToken,
                                modelObj=modelObj, modelType=modelType)
        return rtnValAdd

    '''
        Embedded model Upload function
        Accepts: filenames and file stream as a tuple on model config and model schema
    '''
    def uploadEmbeddingConfig(self, modelConfig: (str, Any) = None, modelSchema: (str, Any) = None, modelType: str = "DNN"):
        rtnValAdd = UploadModelConfig(url=self.url, requestId=self.requestId, authToken=self.authToken,
                                      modelConfig=modelConfig, modelSchema=modelSchema, modelType=modelType)
        return rtnValAdd

    '''
        Upload selected Knowledge Cases (feature sets), that will build the initial Sequential Collection case base
    '''
    def uploadModelCaseData(self, caseDataSet: (str, Any) = None):
        rtnVal = UploadModelCaseData(
            url=self.url,
            requestId=self.requestId,
            authToken=self.authToken,
            caseDataSet = caseDataSet
        )

        return rtnVal

    '''
        Process the upload DNN model with Synapses Logger embedding and dynamically creating the Sequential Collection case base
    '''

    def processEmbeddedModel(self, caseType: str = ""):
        rtnVal = ProcessModel(
            url=self.url,
            requestId=self.requestId,
            authToken=self.authToken,
            refCaseType=caseType
        )

        return rtnVal

    '''
        Foundational Model Profiling function
        Accepts: Case-Type, Network Type, Model Name, Prompt List, Keyword List and the Token Length
    '''
    # Reasoning:Base
    def __GetFMProfilingBase(self, url: str = "", requestId: str = "", authToken: str = "", refCaseType: str = "", nType: int = 0,
                           modelName: str = "", promptTextList: List[str] = [], keyList:List[str] = [],
                           tokenLen:int = 100):
        if url == "":
            url = BASE_URL_API

        result = requests.post(url + '/internal/profiling',
                               json={'requestId': requestId, 'authToken': authToken,
                                     'refCaseType': refCaseType,
                                     'nType': nType, 'modelName': modelName, 'promptTextList': promptTextList,
                                     'keyList': keyList, 'tokenLen': tokenLen})
        return result.status_code, result.reason, result.text

    def generateFMProfiling(self, refCaseType: str = "", nType: int = 2, modelName: str = "meta-llama/Llama-3.2-1B", promptTextList: List[str] = [], keyList:List[str] = [], tokenLen:int = 100):

        rtnVal = self.__GetFMProfilingBase(
            url=self.url,
            requestId=self.requestId,
            authToken=self.authToken,
            refCaseType=refCaseType,
            nType=nType,
            modelName=modelName,
            promptTextList=promptTextList,
            keyList=keyList,
            tokenLen=tokenLen,
        )

        return rtnVal


import json
from typing import Any, List, Tuple
import concurrent.futures
import requests
import base64

from tikos.config import VER, BASE_URL_API

class Tooling(object):

    def __init__(self, url: str = "", requestId: str = "", authToken: str = ""):
        if url == "":
            self.url = BASE_URL_API
        else:
            self.url = url
        self.requestId = requestId
        self.authToken = authToken

    '''
        Foundational Model Profiling Matching tool
        Accepts: PayloadId, Reference Docs, Case-Type, Network Type, Reasoning Type, Similarity Type, NMP model Type, Payload Config, Network Type, Model Name, Prompt List and the Token Length
    '''
    # Reasoning:Base
    def __GetProfileMatchingBase(self, url: str = "", requestId: str = "", authToken: str = "", payloadId: str = "", refdoc: str = "", refCaseType: str = "", RType: str = "", WType: str = "", llmmodel: int = 2, payloadconfig: str = "", nType: int = 0,
                           modelName: str = "", promptTextList: List[str] = [], tokenLen:int = 100):
        if url == "":
            url = BASE_URL_API

        result = requests.post(url + '/tooling/profiling',
                               json={'requestId': requestId, 'authToken': authToken, 'payloadId': payloadId, 'refdoc': refdoc,
                                     'refCaseType': refCaseType, 'RType': RType, 'WType': WType, 'X-TIKOS-MODEL': llmmodel, 'payloadconfig': payloadconfig,
                                     'nType': nType, 'modelName': modelName, 'promptTextList': promptTextList, 'tokenLen': tokenLen})
        return result.status_code, result.reason, result.text

    def generateFMProfileMatching(self, payloadId: str = "", refdoc: str = "", refCaseType: str = "", RType: str = "DEEPCAUSAL_PROFILE_PATTERN_ADV", WType: str = "PROFILING", llmmodel: int = 2, payloadconfig: str = "", nType: int = 2,
                           modelName: str = "meta-llama/Llama-3.2-1B", promptTextList: List[str] = [], tokenLen:int = 100):

        rtnVal = self.__GetProfileMatchingBase(
            url=self.url,
            requestId=self.requestId,
            authToken=self.authToken,
            payloadId=payloadId,
            refdoc=refdoc,
            refCaseType=refCaseType,
            RType=RType,
            WType=WType,
            llmmodel=llmmodel,
            payloadconfig=payloadconfig,
            nType=nType,
            modelName=modelName,
            promptTextList=promptTextList,
            tokenLen=tokenLen,
        )

        return rtnVal

    '''
        Foundational Model Profiling Guard Railing tool
        Accepts: PayloadId, Reference Docs, Case-Type, Network Type, Reasoning Type, Similarity Type, NMP model Type, Payload Config, Network Type, Model Name, Prompt List and the Token Length
    '''
    # Reasoning:Base
    def __GetProfileGuardRailingBase(self, url: str = "", requestId: str = "", authToken: str = "", payloadId: str = "",
                                 refdoc: str = "", refCaseType: str = "", RType: str = "", WType: str = "",
                                 llmmodel: int = 2, payloadconfig: str = "", nType: int = 0,
                                 modelName: str = "", promptTextList: List[str] = [], tokenLen: int = 100):
        if url == "":
            url = BASE_URL_API

        result = requests.post(url + '/tooling/profiling/guardrailing',
                               json={'requestId': requestId, 'authToken': authToken, 'payloadId': payloadId,
                                     'refdoc': refdoc,
                                     'refCaseType': refCaseType, 'RType': RType, 'WType': WType,
                                     'X-TIKOS-MODEL': llmmodel, 'payloadconfig': payloadconfig,
                                     'nType': nType, 'modelName': modelName, 'promptTextList': promptTextList,
                                     'tokenLen': tokenLen})
        return result.status_code, result.reason, result.text

    def generateFMProfileGuardRailing(self, payloadId: str = "", refdoc: str = "", refCaseType: str = "", RType: str = "DEEPCAUSAL_PROFILE_PATTERN_ADV", WType: str = "PROFILING", llmmodel: int = 2, payloadconfig: str = "", nType: int = 2,
                           modelName: str = "meta-llama/Llama-3.2-1B", promptTextList: List[str] = [], tokenLen:int = 100):

        rtnVal = self.__GetProfileGuardRailingBase(
            url=self.url,
            requestId=self.requestId,
            authToken=self.authToken,
            payloadId=payloadId,
            refdoc=refdoc,
            refCaseType=refCaseType,
            RType=RType,
            WType=WType,
            llmmodel=llmmodel,
            payloadconfig=payloadconfig,
            nType=nType,
            modelName=modelName,
            promptTextList=promptTextList,
            tokenLen=tokenLen,
        )

        return rtnVal

    '''
        Deep learning model robustness analysis tool
        Accepts: JSON payload string 
        {
            "models": [
                {
                    "model_name": "Model 1",
                    "case_type": "iris flowers",
                    "target": "species",
                    "org_id": "tikos",
                    "input_features": [
                        "sepal_length",
                        "sepal_width",
                        "petal_length",
                        "petal_width"
                    ],
                    "session": {
                        "requestId": "tikos",
                        "authToken": "tikos"
                    },
                    "model_size": 000,
                    "test_data_filename": "file.csv"
                }
            ]
        }
    '''
    # Model Robustness:Base
    def __analyseModelRobustness(self, url: str = "", payloadJsonStr: str = ""):
        if url == "":
            url = BASE_URL_API

        jsonPayload= json.dumps(payloadJsonStr)

        uploadFiles = {
            'json': ('payload.json', jsonPayload, 'application/json')
        }

        result = requests.post(url + '/tooling/modelrobustness',files=uploadFiles)

        return result.status_code, result.reason, result.text

    def analyseModelRobustness(self, payloadJsonStr: str = ""):

        rtnVal = self.__analyseModelRobustness(
            url=self.url,
            payloadJsonStr=payloadJsonStr)

        return rtnVal


    '''
        Deep learning model Feature Association analysis tool
        Accepts: JSON payload string
        eg: 
        {
            "visualizing_info": {
                "case_type": "cancer_type",
                "target": "net_survival",
                "org_id": "tikos",
                "input_features": [
                    "cancer_type",
                    "survival_type",
                    "stage",
                    "age_group",
                    "sex",
                    "survival_time_years",
                    "number_of_patients"
                ],
                "session": {
                    "requestId": "tikos",
                    "authToken": "tikos"
                },
                "test_data_filename": "file.csv"
            }
        }
    '''
    # Model Feature Association:Base
    def __analyseModelFeatureAssociation(self, url: str = "", payloadJsonStr: str = ""):
        if url == "":
            url = BASE_URL_API

        jsonPayload = json.dumps(payloadJsonStr)

        uploadFiles = {
            'json': ('payload.json', jsonPayload, 'application/json')
        }

        result = requests.post(url + '/tooling/featureassociation', files=uploadFiles)

        return result.status_code, result.reason, result.text

    def analyseModelFeatureAssociation(self, payloadJsonStr: str = ""):

        rtnVal = self.__analyseModelRobustness(
            url=self.url,
            payloadJsonStr=payloadJsonStr)

        return rtnVal

    def analyseModelFeatureAssociationSerialised(self, payloadJsonStr: str = ""):

        statusCode, statusReason, rtnText = self.__analyseModelRobustness(
            url=self.url,
            payloadJsonStr=payloadJsonStr)

        if statusCode == 200:
            jsonLoad = json.loads(rtnText)
            rtnLst = []

            for i, viz in enumerate(jsonLoad.get("visualizations", [])):
                print(f"  Title: {viz.get('title')}")
                base64_string = viz.get('image_base64')
                if base64_string:
                    image_bytes = base64.b64decode(base64_string)

                    filename = f"{viz.get('title', f'image_{i + 1}').replace(' ', '_')}.png"
                    rtnLst.append((filename, image_bytes))
                else:
                    print(" No image data found for this visualization.")

            return rtnLst
        else:
            return None